#ifndef Board_h
#define Board_h

#include "Map.h"
#include "Matrix.h"


macro bool pause(void)      // DEBUGGING
{
    char    buf[100];
    printf("[Pause]"); fflush(stdout);
    fgets(buf, sizeof(buf), stdin);
    return (buf[0] == '\n');
}


//=================================================================================================
// Basic types:


enum Tile {
    Land,
    Base,
    Aqua,
    Wall,
};

macro bool passable(Tile t) { return (unsigned)t <= (unsigned)Base; }

typedef pair<short,short> Pos;

enum Move {
    Idle,
    Pick,
    Drop,
    MoveN, MoveS, MoveE, MoveW,
};

struct Package {
    int     id;         // -1 means no package
    int     weight;
    Pos     pos;        // Will be (-1, id) if carried by robot 'id'.
    Pos     dest;
    bool    dropped;    // True means intentionally dropped here (so may be vanished due to delivery)
    bool    undeliverable;

    Package(void) : id(-1), weight(0), pos(Pos(0,-1)), dest(Pos(0,-1)), dropped(false), undeliverable(false) { }
};

struct Robot {
    int         id;     // -1 means dead (no) robot 
    Pos         pos;
    vec<int>    packs;
    Move        last_move;
    bool        was_pushed;

    Robot(void) : id(-1), pos(Pos(0,-1)), last_move(Idle), was_pushed(false) { }
};


// A direcion is just an 'int'.

#define NoDir -1
#define North 0
#define South 1
#define East  2
#define West  3

macro Move dirToMove(int dir) { return (dir == North) ? MoveN : (dir == South) ? MoveS : (dir == East) ? MoveE : (dir == West) ? MoveW : Idle; }

macro int flipDir  (int dir) { return (dir == NoDir) ? NoDir : (dir ^ 1); }
macro int turnRight(int dir) { return (dir == North) ? East : (dir == East) ? South : (dir == South) ? West : (dir == West) ? North : NoDir; }
macro int turnLeft (int dir) { return (dir == North) ? West : (dir == West) ? South : (dir == South) ? East : (dir == East) ? North : NoDir; }

macro const char* dirName(int dir) { return (dir == North) ? "North" : (dir == West) ? "West" : (dir == South) ? "South" : (dir == East) ? "East" : "NoDir"; }


//=================================================================================================
// Missions:


enum Mission_type {
    NoMission,
    CheckBase,
    Fetch,
    Deliver,
    Harass
};


struct Mission {
    Mission_type    type;
    double          value;      // Evaluation based on profit vs risk.
    Pos             goal;

    Mission(void) : type(NoMission), value(0), goal(Pos(0,-1)) { }
    Mission(Mission_type t, double v, Pos p) : type(t), value(v), goal(p) { }

    bool operator < (const Mission& other) const { return value > other.value; }    // (reverse order)

    // Debug:
    void dump(void);
};


struct HPiece {     // (piece of history)
    double  dist;
    Pos     pos;
    Move    action;
    int     bet;
    Move    last_move;
    bool    was_pushed;
};


struct MissionControl {
    Mission current;
    double  estimated_cost;
    double  cost_so_far;

    // Banning:
    Map<Pos,float>      banned;             // How banned? Value of mission will be multiplied with 'F^ban' where F is a penalty factor.
    vec<bool>           undeliverables;     // A package was detected as undeliverable. It has no value.
    vec<pair<Pos,int> > ban_removal;        // Counts down one each move. 'banned' is decrease by 1 when counter reach zero.

    void ban    (Pos p, double v)        { banned.set(p, banned.at(p) + v); }
    void banTemp(Pos p, double v, int c) { banned.set(p, banned.at(p) + v); ban_removal.push(pair_new(p,c)); }
    void updateTempBans(void);

    // Shortest path:
    Matrix<float> dist;     // (updated by 'plan()')
    Matrix<float> cost;     // (updated by 'plan()')
    Matrix<float> fee;      // (updated by 'plan()') -- used to compute cost 

    // History:
    vec<HPiece> old;

    // trail fr egen robot -- undvik rutor. nolla vid ny mission.

    // Debug:
    vec<Mission>    last_missions;
};

inline void MissionControl::updateTempBans(void) {
    for (int i = 0; i < ban_removal.size();){
        if (ban_removal[i].snd == 0){
            ban(ban_removal[i].fst, -1);
            ban_removal[i] = ban_removal.last();
            ban_removal.pop();
        }else{
            ban_removal[i].snd--;
            i++;
        }
    }
}


//=================================================================================================
// Board (whole game):


struct Board {
    // Translate id's:      // Not implemeted! Hope it is not necessary!
    //Map<int,int>    to_internal;
    //vec<int>        from_internal;

    // Map:
    Matrix<Tile>    tile;       // Tile at position.

    int     height(void) { return tile.height(); }
    int     width (void) { return tile.width (); }

    // Robots and packages: 
    vec<Package>    packinfo;
    vec<Robot>      roboinfo;
    vec<Pos>        bases;      // Only unvisited bases are listed here

    void move (int id, int delta_x, int delta_y);
    void pick (int id, int pkg_id);
    void drop (int id, int pkg_id);
    void enter(int id, int x, int y);
    void die  (int id);

    // Me:
    int     my_id;      // My ID
    int     capacity;   // How much can I carry (constant)
    int     money;      // Money left to spend (will shrink)
    int     score;      // My current score
    bool    first_move; // First move must be handled specially (lots of pushing)

    Robot&  me() { return roboinfo[my_id]; }

    // Stats:
    int     initial_nof_bases;

    // Mission control:
    MissionControl  mc;

    // Debug:
    void dump(bool ansi, bool big, bool clear);
};


//=================================================================================================

#endif
